"""Classes and functions to facilitate local parallel computing."""

import multiprocessing as mp

mp.set_start_method("fork")
CPUS = mp.cpu_count()


class ProgressBar:
    """
    Display the degree of progress for naturally iterative jobs (for
    example while parallel processing over a large work load).
    """
    def __init__(self, current, end, width=50):
        """
        Initialize an instance of ProgressBar

        :param current: current index of work unit
        :type current: int
        :param end: total length of work units
        :type end: int
        :param width: desired character width of ProgressBar
        :type width: int
        """
        self._pct = int(float(current) / end * 100)
        self._char_width = int(100 // width)
        self._fill = self._pct // self._char_width
        self._pad = width - self._fill

    def __str__(self):
        return f"[{'#' * self._fill}{' ' * self._pad}] {self._pct}%"


def show_progress(current, end, width=50):
    """
    Command-line updating progressbar

    :param current: current index of work unit
    :type current: int
    :param end: total length of work units
    :type end: int
    :param width: desired character width of ProgressBar
    :type width: int
    :return: progress
    """
    progress = ProgressBar(current, end, width)
    print(f"\r{progress}", end="")
    return progress


def parallelize(inputs, num_processors, task, verbose=False):
    """
    Parallelize some task on an input list across the specified number
    of processors
    :param inputs: list of inputs
    :param num_processors: number of processor cores to use
    :param task: name of the function to run
    :param verbose: updating progress bar output?
    :return: results
    """
    results = []

    # Don't do any work if there are no inputs
    if len(inputs) == 0:
        return results

    num_processors = count_processors(inputs, num_processors)

    tasks = []
    for item in inputs:
        if not isinstance(item, tuple):
            item = (item,)
        tasks.append((task, item))

    # Start working on the jobs
    results = start_processes(tasks, num_processors, verbose)

    return results


def count_processors(inputs, num_processors):
    """
    Programmatically determines whether the specified num_processors is
    appropriate. There's no need to use more processors than there are
    inputs, and it's impossible to use fewer than 1 processor or more
    than exist on the machine running the code.
    :param inputs: list of inputs
    :param num_processors: specified number of processors
    :return: num_processors (optimized)
    """
    if num_processors < 1 or num_processors > CPUS:
        print(f"Invalid number of CPUs specified ({num_processors})")
        num_processors = min([mp.cpu_count(), len(inputs)])
        print(f"Using {num_processors} CPUs...")

    return num_processors


def worker(input_queue, output_queue):
    for func, args in iter(input_queue.get, 'STOP'):
        result = func(*args)
        output_queue.put(result)
    return


def start_processes(inputs, num_processors, verbose):
    """
    Creates input and output queues, and runs the jobs
    :param inputs: jobs to run
    :param num_processors: optimized number of processors
    :param verbose: updating progress bar output?
    :return: results
    """
    job_queue = mp.Queue()
    done_queue = mp.Queue()

    # Counter so we know how many tasks in all (input + show_progress tasks)
    tasks = 0

    if verbose is True:
        interval = max([1, len(inputs)//100])

        # Put inputs into job queue
        for i, task in enumerate(inputs):
            tasks += 1
            if i % interval == 0:
                job_queue.put((show_progress, (i, len(inputs))))
                tasks += 1
            job_queue.put(task)
        tasks += 1
        job_queue.put((show_progress, (len(inputs), len(inputs))))
    else:
        for i, task in enumerate(inputs):
            tasks += 1
            job_queue.put(task)

    # Put a bunch of 'STOP' signals at the end of the queue
    for i in range(num_processors):
        job_queue.put('STOP')

    # Start up workers
    worker_pool = []
    for i in range(num_processors):
        worker_n = mp.Process(target=worker, args=(job_queue, done_queue))
        worker_n.start()
        worker_pool.append(worker_n)

    # Grab results from done queue
    results = []

    # Remove non-Progress results
    for i in range(tasks):
        result = done_queue.get()
        if not isinstance(result, ProgressBar):
            results.append(result)

    [worker_n.join() for worker_n in worker_pool]

    # Leave the progress bar line
    if verbose is True:
        print("")

    return results



from phamerate.cmd import run_command
from phamerate.fasta import parse_fasta, write_fasta


def clustalo_align(fasta, align):
    """
    Use clustalo to align the sequences in `fasta` and write the
    alignment to `align`.

    :param fasta: path to a FASTA multiple sequence file
    :type fasta: pathlib.Path
    :param align: path to a FASTA multiple sequence alignment file
    :type align: pathlib.Path
    """
    command = f"clustalo -i {fasta} --infmt=fasta -o {align} " \
              f"--outfmt=fasta --output-order=tree-order --threads=1 " \
              f"--seqtype=protein --force"
    run_command(command)


def align_phams(phams, fasta_paths, align_paths):
    """
    Align a pham using Clustal Omega

    :param phams: the phams to align
    :type phams: list of classes.Pham
    :param fasta_paths: paths to FASTA files to align
    :type fasta_paths: list of pathlib.Path
    :param align_paths: paths where the alignments should go
    :type align_paths: list of pathlib.Path
    """
    for pham, fasta, align in zip(phams, fasta_paths, align_paths):
        # Only align phams with > 1 non-redundant sequence
        if len([x for x in pham.nr_iterator]) != 1:
            clustalo_align(fasta, align)

            # Make sure all this pham's headers are in the alignment file
            _, aligned_translations = parse_fasta(align)
            for aligned_translation in aligned_translations:
                pham.add_aligned_translation(aligned_translation)

            headers, sequences = list(), list()
            for _, geneid, translation in pham.alignment_iterator:
                headers.append(geneid), sequences.append(translation)
            write_fasta(headers, sequences, align)

        # Make sure all this pham's headers are in the FASTA file
        headers, sequences = list(), list()
        for _, geneid, translation in pham:
            headers.append(geneid), sequences.append(translation)
        write_fasta(headers, sequences, fasta)



"""Some classes that serve useful functions..."""


class SequenceDB:
    """
    Provide a convenient interface for wrangling translations and
    gene identifiers from different sources
    """
    def __init__(self):
        """
        Initialize an instance of SequenceDB.
        """
        self._translations_to_geneids = dict()
        self._geneids_to_translations = dict()
        self._genomes_to_geneids = dict()
        self._geneids_to_genomes = dict()

    @property
    def translations(self):
        """
        Provide read-only iterable access to the translations
        represented in this database.

        :return: translations
        """
        return self._translations_to_geneids.keys()

    @property
    def genomes(self):
        """
        Provide read-only iterable access to the genomes represented in
        this database.

        :return: genomes
        """
        return self._genomes_to_geneids.keys()

    @property
    def nr_iterator(self):
        """
        Alternative to __iter__() that allows iterating over non-redundant
        sequences rather than all sequences.

        :rtype: tuple[str, str]
        """
        for translation, geneids in self._translations_to_geneids.items():
            yield geneids[0], translation

    def add_gene(self, geneid, genome, translation):
        """
        Add a gene to the database.

        :param geneid: the gene's identifier
        :type geneid: str
        :param genome: the genome the gene comes from
        :type genome: str
        :param translation: the gene's protein sequence
        :type translation: str
        """
        if geneid in self._geneids_to_translations:
            raise ValueError(f"geneid {geneid} already in database")

        self._geneids_to_translations[geneid] = translation

        if translation in self._translations_to_geneids:
            self._translations_to_geneids[translation].append(geneid)
        else:
            self._translations_to_geneids[translation] = [geneid]

        self._geneids_to_genomes[geneid] = genome

        if genome in self._genomes_to_geneids:
            self._genomes_to_geneids[genome].append(geneid)
        else:
            self._genomes_to_geneids[genome] = [geneid]

    def get_translation_geneids(self, translation):
        """
        Lookup the indicated translation and return its associated
        geneids.

        :param translation: the translation to lookup
        :type translation: str
        :return: geneids
        :rtype: list[str]
        :raise: KeyError
        """
        return self._translations_to_geneids[translation]

    def get_geneid_translation(self, geneid):
        """
        Lookup the indicated geneid and return its translation.

        :param geneid: the geneid to lookup
        :type geneid: str
        :return: translation
        :rtype: str
        :raise: KeyError
        """
        return self._geneids_to_translations[geneid]

    def get_genome_geneids(self, genome):
        """
        Lookup the indicated genome and return its associated geneids.

        :param genome: the genome to lookup
        :type genome: str
        :return: geneids
        :rtype: list[str]
        :raise: KeyError
        """
        return self._genomes_to_geneids[genome]

    def get_geneid_genome(self, geneid):
        """
        Lookup the indicated geneid and return its genome.

        :param geneid: the geneid to lookup
        :type geneid: str
        :return: genome
        :rtype: str
        :raise: KeyError
        """
        return self._geneids_to_genomes[geneid]

    def get_genome_translations(self, genome):
        """
        Lookup the indicated genome and return its associated
        translations.

        :param genome: the genome to lookup
        :type genome: str
        :return: translations
        :rtype: list[str]
        :raise: KeyError
        """
        return [self.get_geneid_translation(x) for x in
                self.get_genome_geneids(genome)]

    def __iter__(self):
        for geneid, translation in self._geneids_to_translations.items():
            genome = self.get_geneid_genome(geneid)
            yield genome, geneid, translation

    def __len__(self):
        return len(self._geneids_to_translations)


class Pham(SequenceDB):
    """
    A type of SequenceDB, with additional convenience methods for
    performing actions that are useful when considering relatively
    small numbers of related sequences, and not so much for whole
    meta- or pan-genomes.
    """
    def __init__(self):
        super().__init__()
        self._geneids_to_aligned_translations = dict()

    @property
    def alignment_iterator(self):
        """
        Alternative to __iter__(), which allows iteration over aligned
        sequences rather than the unaligned sequences.
        """
        alignments = self._geneids_to_aligned_translations.items()
        for geneid, translation in alignments:
            genome = self.get_geneid_genome(geneid)
            yield genome, geneid, translation

    def merge(self, other):
        """
        Merge phams from multiple rounds of clustering.

        :param other: another pham to be merged into this one
        :type other: Pham
        """
        if not isinstance(other, Pham):
            raise TypeError(f"cannot merge Pham with {type(other)}")

        for genome, geneid, translation in other:
            try:
                self.add_gene(geneid, genome, translation)
            except ValueError:
                pass    # Tried adding a gene that was already in the pham

    def add_aligned_translation(self, aligned_translation):
        """
        Add an aligned translation for this database to keep track of.

        :param aligned_translation: an aligned translation
        :type aligned_translation: str
        :raise: KeyError
        """
        translation = aligned_translation.replace("-", "")
        geneids = self.get_translation_geneids(translation)
        for geneid in geneids:
            self._geneids_to_aligned_translations[geneid] = aligned_translation

    def in_genome(self, genome):
        """
        Check whether this phamily is present in the indicated genome.

        :param genome: the genome to be checked against
        :type genome: str
        return: in_genome
        """
        return genome in self.genomes

    def minimum_length(self, kind="nucleotide"):
        """
        Return the length of the shortest sequence in the phamily. Can
        return either 'nucleotide' or 'translation' length.

        :param kind: the type of sequence length to return
        :type kind: str
        :return: min_length
        """
        if kind not in ("nucleotide", "translation"):
            raise ValueError(f"argument 'kind' must be one of ('nucleotide', "
                             f"'translation')")

        lengths = [len(x) for x in self.translations]
        min_length = min(lengths)

        if kind == "nucleotide":
            return min_length * 3 + 3

        return min_length

    def average_length(self, kind="nucleotide"):
        """
        Return the length of the average sequence in the phamily. Can
        return either 'nucleotide' or 'translation' length.

        :param kind: the type of sequence length to return
        :type kind: str
        :return: avg_length
        """
        if kind not in ("nucleotide", "translation"):
            raise ValueError(f"argument 'kind' must be one of ('nucleotide', "
                             f"'translation')")

        lengths = [len(x) for x in self.translations]
        avg_length = float(sum(lengths))/len(lengths)

        if kind == "nucleotide":
            return avg_length * 3 + 3

        return avg_length

    def maximum_length(self, kind="nucleotide"):
        """
        Return the length of the longest sequence in the phamily. Can
        return either 'nucleotide' or 'translation' length.

        :param kind: the type of sequence length to return
        :type kind: str
        :return: max_length
        """
        if kind not in ("nucleotide", "translation"):
            raise ValueError(f"argument 'kind' must be one of ('nucleotide', "
                             f"'translation')")

        lengths = [len(x) for x in self.translations]
        max_length = max(lengths)

        if kind == "nucleotide":
            return max_length * 3 + 3

        return max_length

    def __lt__(self, other):
        if not isinstance(other, Pham):
            raise TypeError(f"cannot merge Pham with {type(other)}")

        return len(self) < len(other)



"""Functions for writing and parsing FASTA files."""


def write_fasta(headers, sequences, filepath, width=80):
    """
    Write the given headers and sequences to filepath, wrapping long
    sequence lines at the indicated width.

    :param headers: sequence labels
    :type headers: str or list of str
    :param sequences: nucleotide or amino acid sequences
    :type sequences: str or list of str
    :param filepath: a FASTA file to write to
    :type filepath: pathlib.Path or str
    :param width: maximum number of characters per sequence line
    :type width: int
    :return: filepath
    """
    if not isinstance(headers, list) or not isinstance(sequences, list):
        raise TypeError(f"headers and sequences should be lists of strings")

    # Open the file for writing
    fasta_writer = open(filepath, "w")

    for header, sequence in zip(headers, sequences):
        fasta_writer.write(f">{header}\n")
        # Use string slicing to split the sequence to satisfy the width param
        for i in range(0, len(sequence), width):
            fasta_writer.write(f"{sequence[i:i + width]}\n")

    # Close the file
    fasta_writer.close()

    return filepath


def parse_fasta(filepath):
    """
    Parse a FASTA file and return the headers and sequences

    :param filepath: a FASTA file to parse
    :type filepath: pathlib.Path or str
    :return: headers, sequences
    """
    headers, sequences = list(), list()

    # Open the file for reading
    fasta_reader = open(filepath, "r")

    cache = list()
    for line in fasta_reader:
        # If the line is a header line, flush the cache and store new header
        if line.startswith(">"):
            sequences.append("".join(cache))
            cache = list()
            headers.append(line.lstrip(">").rstrip())
        # Otherwise, append to the cache
        else:
            cache.append(line.rstrip())
    # Flush the last sequence out of the cache, and pop the empty sequence
    sequences.append("".join(cache))
    sequences = sequences[1:]

    # Close the file
    fasta_reader.close()

    return headers, sequences



"""Functions for performing pham assembly with MMseqs2."""

from phamerate.cmd import run_command


def mmseqs_createdb(fasta, tmp_dir):
    """
    Run 'mmseqs createdb' command.

    :param fasta: path to the FASTA file to convert
    :type fasta: pathlib.Path
    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    """
    command = f"mmseqs createdb {fasta} {tmp_dir}/sequenceDB -v 1"
    run_command(command)


def mmseqs_cluster(tmp_dir, cluster_mode, sens, identity, coverage, evalue,
                   max_seqs=1000, steps=1, align_mode=3, cov_mode=0, threads=1):
    """
    Run 'mmseqs cluster' command.

    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param identity: percent identity threshold for clustering
    :type identity: float
    :param coverage: percent coverage threshold for clustering
    :type coverage: float
    :param evalue: significance threshold for clustering
    :type evalue: float
    :param sens: sensitivity parameter for seeding alignments
    :type sens: float
    :param cluster_mode: clustering algorithm to use
    :type cluster_mode: int
    :param max_seqs: number of hits to store per query
    :type max_seqs: int
    :param steps: number of clustering steps to proceed with
    :type steps: int
    :param align_mode: what kind of alignments to generate
    :type align_mode: int
    :param cov_mode: coverage calculation to use
    :type cov_mode: int
    :param threads: how many threads to use?
    :type threads: int
    """
    command = f"mmseqs cluster {tmp_dir}/sequenceDB {tmp_dir}/clusterDB " \
              f"{tmp_dir} --min-seq-id {identity} -c {coverage} -e {evalue} " \
              f"-s {sens} --max-seqs {max_seqs} --cluster-steps {steps} " \
              f"--threads {threads} --alignment-mode {align_mode} --cov-mode " \
              f"{cov_mode} --cluster-mode {cluster_mode} --cluster-reassign " \
              f"-v 1"
    run_command(command)


def mmseqs_result2profile(tmp_dir, threads=1):
    """
    Run 'mmseqs result2profile' command.

    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param threads: how many threads to use?
    :type threads: int
    :return: command
    """
    command = f"mmseqs result2profile {tmp_dir}/sequenceDB " \
              f"{tmp_dir}/sequenceDB {tmp_dir}/clusterDB {tmp_dir}/profileDB " \
              f"--threads {threads} -v 1"
    run_command(command)


def mmseqs_profile2consensus(tmp_dir, threads=1):
    """
    Run 'mmseqs profile2consensus' command.

    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param threads: how many threads to use?
    :type threads: int
    :return: command
    """
    command = f"mmseqs profile2consensus {tmp_dir}/profileDB " \
              f"{tmp_dir}/consensusDB --threads {threads} -v 1"
    run_command(command)


def mmseqs_search(tmp_dir, identity, coverage, evalue, max_seqs=1000,
                  align_mode=3, cov_mode=0, threads=1):
    """
    Return 'mmseqs search' command.

    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param threads: how many threads to use?
    :type threads: int
    :param identity: percent identity threshold
    :type identity: float
    :param coverage: percent coverage threshold
    :type coverage: float
    :param evalue: significance threshold
    :type evalue: float
    :param threads: number of threads to use
    :type threads: int
    :param max_seqs: maximum number of hits to store per query
    :type max_seqs: int
    :param align_mode: MMseqs2 alignment mode
    :type align_mode: int
    :param cov_mode: MMseqs2 coverage mode
    :type cov_mode: int
    :return: command
    """
    command = f"mmseqs search {tmp_dir}/profileDB {tmp_dir}/consensusDB " \
              f"{tmp_dir}/alignDB {tmp_dir} --min-seq-id {identity} -c " \
              f"{coverage} --cov  {coverage} -e {evalue} --e-profile " \
              f"{evalue} --threads  {threads} --max-seqs {max_seqs} " \
              f"--alignment-mode {align_mode} --cov-mode {cov_mode} " \
              f"--add-self-matches -v 1"
    run_command(command)


def mmseqs_clust(tmp_dir, cluster_mode, threads=1):
    """
    Return 'mmseqs clust' command.

    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param cluster_mode: which clustering algorithm to use?
    :type cluster_mode: int
    :param threads: how many threads to use?
    :type threads: int
    """
    command = f"mmseqs clust {tmp_dir}/consensusDB {tmp_dir}/alignDB " \
              f"{tmp_dir}/resultDB --cluster-mode {cluster_mode} " \
              f"--threads {threads} -v 1"
    run_command(command)


def mmseqs_first_iter_cleanup(tmp_dir, outfile, threads=1):
    """
    Run 'mmseqs createseqfiledb' command after sequence-sequence
    clustering.

    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param outfile: file where phams should be written
    :type outfile: pathlib.Path
    :param threads: how many threads to use?
    :type threads: int
    """
    command = f"mmseqs createseqfiledb {tmp_dir}/sequenceDB " \
              f"{tmp_dir}/clusterDB {tmp_dir}/firstIterDB --threads " \
              f"{threads} -v 1"
    run_command(command)

    command = f"mmseqs result2flat {tmp_dir}/sequenceDB {tmp_dir}/sequenceDB " \
              f"{tmp_dir}/firstIterDB {outfile} -v 1"
    run_command(command)


def mmseqs_second_iter_cleanup(tmp_dir, outfile, threads=1):
    """
    Run 'mmseqs createseqfiledb' command after consensus-HMM
    clustering.

    :param tmp_dir: directory MMseqs2 can use as scratch space
    :type tmp_dir: pathlib.Path
    :param outfile: file where phams should be written
    :type outfile: pathlib.Path
    :param threads: how many threads to use?
    :type threads: int
    """
    command = f"mmseqs createseqfiledb {tmp_dir}/sequenceDB " \
              f"{tmp_dir}/resultDB {tmp_dir}/secondIterDB --threads " \
              f"{threads} -v 1"
    run_command(command)

    command = f"mmseqs result2flat {tmp_dir}/profileDB {tmp_dir}/consensusDB " \
              f"{tmp_dir}/secondIterDB {outfile} -v 1"
    run_command(command)


def parse_mmseqs_output(outfile):
    """
    Parse the indicated MMseqs2 FASTA-like file into a dictionary of
    integer-named phams.

    :param outfile: FASTA-like parseable output
    :type outfile: pathlib.Path
    :return: phams
    :rtype: dict
    """
    phams = {}
    pham_translations = list()
    pham_name = -1

    with open(f"{outfile}", "r") as fh:
        prior = fh.readline()
        current = fh.readline()

        # While loop to iterate until EOF
        while current:
            # If current and previous lines are header lines, new pham block
            if current.startswith(">") and prior.startswith(">"):
                phams[pham_name] = pham_translations
                pham_name += 1
                pham_translations = []

            # If the current line is a translation line, keep growing pham
            if not current.startswith(">"):
                pham_translations.append(current.rstrip())

            prior, current = current, fh.readline()

        # Need to dump the last pham into the dictionary
        phams[pham_name] = pham_translations

    # Pham -1 is a placeholder
    phams.pop(-1)

    return phams



from subprocess import Popen, PIPE, DEVNULL
import shlex


def run_command(command, verbose=False):
    """
    Run the indicated command as a subprocess.

    :param command: the command to run
    :type command: str
    :param verbose: want stdout/stderr?
    :type verbose: bool
    :return: out, err
    """
    command = shlex.split(command)
    if verbose:
        with Popen(command, stdout=PIPE, stderr=PIPE) as sp:
            out = sp.stdout.read().decode("utf-8")
            err = sp.stderr.read().decode("utf-8")
            return out, err
    else:
        with Popen(command, stdout=DEVNULL, stderr=DEVNULL) as sp:
            sp.wait()



#! /usr/bin/env python3

"""Assort phage protein sequences into phamilies using MMseqs2."""

import sys
import argparse
import pathlib
import shutil

from phamerate.clustalo import align_phams
from phamerate.parallelize import CPUS, parallelize
from phamerate.utility import *

# Defaults parameters
TMP_DIR = pathlib.Path("/tmp/phamerate")

CM, S = 0, 4.0                  # --cluster-mode, -s
M, C, E = 0.3, 0.85, 0.001      # --min-seq-id, -c, -e
HM, HC, HE = 0.25, 0.5, 0.001   # --min-seq-id, -c/--cov, -e/--e-profile

EPILOG = """
Steinegger M. and Soeding J. MMseqs2 enables sensitive protein
sequence searching for the analysis of massive data sets. Nature
Biotechnology, 2017. doi: 10.1038/nbt.3988"""


def parse_args():
    """
    Parse command line arguments

    :return: parsed_args
    """
    p = argparse.ArgumentParser(description=__doc__, prog="phamerate",
                                epilog=EPILOG)

    p.add_argument("infile", nargs="+", type=pathlib.Path,
                   help="path to input file(s) in FASTA format")

    mmseqs_args = p.add_argument_group("mmseqs arguments")
    mmseqs_args.add_argument("--cluster-mode", type=int,
                             default=CM, metavar='',
                             help=f"clustering algorithm [default: {CM}]")
    mmseqs_args.add_argument("--sensitivity", type=float,
                             default=S, metavar='',
                             help=f"sensitivity: 1.0 favors speed, 7.5 favors "
                                  f"sensitivity [default: {S}]")
    mmseqs_args.add_argument("--identity", type=float,
                             default=M, metavar='',
                             help=f"percent identity for sequence-sequence "
                                  f"clustering [default: {M}]")
    mmseqs_args.add_argument("--coverage", type=float,
                             default=C, metavar='',
                             help=f"percent coverage for sequence-sequence "
                                  f"clustering [default: {C}]")
    mmseqs_args.add_argument("--evalue", type=float,
                             default=E, metavar='',
                             help=f"E-value threshold for sequence-sequence "
                                  f"clustering [default: {E}]")
    mmseqs_args.add_argument("--no-hmm", action="store_true",
                             help="skip HMM clustering")
    mmseqs_args.add_argument("--hmm-identity", type=float,
                             default=HM, metavar='',
                             help=f"percent identity for consensus-HMM "
                                  f"clustering [default: {HM}]")
    mmseqs_args.add_argument("--hmm-coverage", type=float,
                             default=HC, metavar='',
                             help=f"percent coverage for consensus-HMM "
                                  f"clustering [default: {HC}]")
    mmseqs_args.add_argument("--hmm-evalue", type=float,
                             default=HE, metavar='',
                             help=f"E-value threshold for consensus-HMM "
                                  f"clustering [default: {HE}]")

    p.add_argument("-c", "--cpus", type=int, default=CPUS, metavar='',
                   help=f"number of threads to use [default: {CPUS}]")
    p.add_argument("-v", "--verbose", action="store_true",
                   help="print progress messages")
    p.add_argument("-o", "--outdir", default=pathlib.Path().cwd(),
                   type=pathlib.Path, metavar='',
                   help=f"path to directory where output files should go "
                        f"[default: {str(pathlib.Path().cwd())}]")
    p.add_argument("-t", "--tmpdir", default=TMP_DIR,
                   type=pathlib.Path, metavar='',
                   help=f"path where temporary file I/O should occur "
                        f"[default: {str(TMP_DIR)}]")
    p.add_argument("-a", "--align-phams", action="store_true",
                   help="use Clustal Omega to align phams (this could take "
                        "awhile...)")

    return p.parse_args()


def main():
    args = parse_args()

    infiles = args.infile

    outdir = args.outdir
    if not outdir.is_dir():
        outdir.mkdir(parents=True)

    tmpdir = args.tmpdir
    if tmpdir.is_dir():
        shutil.rmtree(tmpdir)
    tmpdir.mkdir(parents=True)

    verbose = args.verbose
    cpus = args.cpus
    align = args.align_phams

    # Set up MMseqs2 parameters
    first_iter_params = (args.cluster_mode, args.sensitivity, args.identity,
                         args.coverage, args.evalue)
    second_iter_params = (args.cluster_mode, args.sensitivity,
                          args.hmm_identity, args.hmm_coverage, args.hmm_evalue)

    # Read the translations in from files
    database = create_database(infiles)

    if verbose:
        print(f"Found {len(database)} translations in {len(infiles)} files...")

    # Perform pham assembly
    phams = assemble_phams(database, first_iter_params, second_iter_params,
                           verbose=verbose, threads=cpus,
                           skip_hmm=args.no_hmm, tmp_dir=tmpdir)

    if verbose:
        print(f"Found {len(phams)} phamilies in dataset...")

    fasta_dir = outdir.joinpath("phamily_fastas")
    if not fasta_dir.is_dir():
        fasta_dir.mkdir()

    write_pham_fastas(phams, fasta_dir, nr_only=align)

    if align:
        if verbose:
            print(f"Computing phamily alignments with Clustal Omega...")

        align_dir = outdir.joinpath("phamily_aligns")
        if not align_dir.is_dir():
            align_dir.mkdir(parents=False)

        # Create 100 jobs so each job increments ProgressBar by 1%
        pre_jobs, jobs = dict(), list()
        for i in range(len(phams)):
            job_key = i % 100

            fasta_path = fasta_dir.joinpath(f"pham_{i+1}.faa")
            align_path = align_dir.joinpath(f"pham_{i+1}.aln")

            if job_key in pre_jobs:
                pre_jobs[job_key]["fastas"].append(fasta_path)
                pre_jobs[job_key]["aligns"].append(align_path)
                pre_jobs[job_key]["phams"].append(phams[i])
            else:
                pre_jobs[job_key] = {"fastas": [fasta_path],
                                     "aligns": [align_path],
                                     "phams": [phams[i]]}

        for job_dict in pre_jobs.values():
            jobs.append((job_dict["phams"], job_dict["fastas"],
                         job_dict["aligns"]))
        parallelize(jobs, cpus, align_phams, verbose)

    if verbose:
        print("Performing basic pan/meta-genome analyses...")

    # Roary-style pangenome analyses
    pangenome_map = map_pangenome(phams, len(database.genomes))
    summarize(pangenome_map, phams, list(database.genomes), outdir)

    if verbose:
        print("Done!")


if __name__ == "__main__":
    # Invoke the help menu if phamerate was run without args
    if len(sys.argv) == 1:
        sys.argv.append("-h")
    main()



from Bio import SeqIO


def parse_genbank(filepath):
    """
    Parse CDS features from records in a Genbank flatfile.

    :param filepath:
    :return: headers, sequences
    """
    headers, sequences = list(), list()
    for record in SeqIO.parse(filepath, "genbank"):
        for feature in record.features:
            if feature.type == "CDS":
                headers.append(feature.qualifiers["locus_tag"][0])
                sequences.append(feature.qualifiers["translation"][0])

    return headers, sequences



"""High-level utility functions for pham assembly and post-processing."""

from phamerate.classes import SequenceDB, Pham
from phamerate.fasta import parse_fasta, write_fasta
from phamerate.mmseqs import *

HEADER = ["Gene", "Non unique Gene name", "No. isolates", "No. sequences",
          "Avg sequences per isolate", "Genome Fragment",
          "Order within Fragment", "Accessory Fragment",
          "Accessory Order with Fragment", "QC", "Min group size nuc",
          "Max group size nuc", "Avg group size nuc"]


def create_database(input_files):
    """
    Walk a list of FASTA files and map non-redundant translations
    to the genomes and labels associated with them.

    :param input_files: FASTA file(s) to map translations from
    :type input_files: list of pathlib.Path
    :return: dataset
    """
    database = SequenceDB()
    for input_file in input_files:
        genome = input_file.stem
        headers, sequences = parse_fasta(input_file)
        for geneid, translation in zip(headers, sequences):
            database.add_gene(geneid, genome, translation)

    return database


def dump_nr_translations(database, filepath):
    """
    Dump the non-redundant translations to a FASTA file for clustering.

    :param database: a database containing genes
    :type database: SequenceDB or Pham
    :param filepath: file to write non-redundant translations to
    :type filepath: pathlib.Path
    """
    headers, sequences = list(), list()
    for geneid, translation in database.nr_iterator:
        headers.append(geneid)
        sequences.append(translation)
    write_fasta(headers, sequences, filepath)


def merge_pre_and_hmm_phams(hmm_phams, pre_phams, database):
    """
    Merge first and second iteration phams.

    :param hmm_phams: clustered consensus sequences
    :type hmm_phams: dict
    :param pre_phams: clustered sequences (used to generate hmms)
    :type pre_phams: dict
    :param database: the database whose sequences were clustered
    :type database: SequenceDB
    :return: merged_phams
    """
    phams, lookup = list(), dict()
    for key, translations in pre_phams.items():
        pham = Pham()
        for translation in translations:
            lookup[translation] = key
            for geneid in database.get_translation_geneids(translation):
                genome = database.get_geneid_genome(geneid)
                pham.add_gene(geneid, genome, translation)
        phams.append(pham)

    if not hmm_phams:
        return sorted(phams, reverse=True)

    merged_phams = list()
    for key, translations in hmm_phams.items():
        pham = Pham()
        for translation in translations:
            pre_pham_key = lookup[translation]
            pham.merge(phams[pre_pham_key])
        merged_phams.append(pham)

    return sorted(merged_phams, reverse=True)


def assemble_phams(database, first_iter, second_iter, threads,
                   verbose, skip_hmm, tmp_dir):
    """
    Perform pham assembly and return the resultant phams.

    :param database: the sequence database
    :type database: SequenceDB
    :param first_iter: sequence-sequence clustering parameters
    :param first_iter: tuple of float
    :param second_iter: consensus-HMM clustering parameters
    :type second_iter: tuple of float
    :param threads: number of threads to use
    :type threads: int
    :param verbose: print messages along the way?
    :type verbose: bool
    :param skip_hmm: skip HMM clustering?
    :type skip_hmm: bool
    :param tmp_dir: temporary directory
    :type tmp_dir: pathlib.Path
    :return: phams
    """
    if verbose:
        print("Creating MMseqs2 database...")

    nr_fasta = tmp_dir.joinpath("all_genes.fasta")
    iter_1_out = tmp_dir.joinpath("firstIter.txt")
    iter_2_out = tmp_dir.joinpath("secondIter.txt")

    dump_nr_translations(database, nr_fasta)
    mmseqs_createdb(nr_fasta, tmp_dir)

    if verbose:
        print("Performing sequence-sequence clustering...")

    clu_mode, sens, i, c, e = first_iter
    mmseqs_cluster(tmp_dir, clu_mode, sens, i, c, e, threads=threads)

    if verbose:
        print("Parsing first iteration phams...")

    mmseqs_first_iter_cleanup(tmp_dir, iter_1_out, threads)

    first_iter_phams = parse_mmseqs_output(iter_1_out)

    if not skip_hmm:
        if verbose:
            print("Building HMMs from pre-phams...")

        mmseqs_result2profile(tmp_dir, threads)

        if verbose:
            print("Extracting consensus sequences from HMMs...")

        mmseqs_profile2consensus(tmp_dir, threads)

        if verbose:
            print("Performing consensus-HMM clustering...")

        clu_mode, sens, i, c, e = second_iter
        mmseqs_search(tmp_dir, i, c, e, threads=threads)

        mmseqs_clust(tmp_dir, clu_mode, threads)

        if verbose:
            print("Parsing second iteration phams...")

        mmseqs_second_iter_cleanup(tmp_dir, iter_2_out, threads)

        second_iter_phams = parse_mmseqs_output(iter_2_out)
    else:
        second_iter_phams = None

    return merge_pre_and_hmm_phams(second_iter_phams, first_iter_phams,
                                   database)


def write_pham_fastas(phams, fasta_dir, nr_only=False):
    """
    Dump pham sequences to FASTA files (one file per pham).

    :param phams: the assembled phams
    :type phams: list[Pham]
    :param fasta_dir: the directory where files should be written
    :type fasta_dir: pathlib.Path
    :param nr_only: only write non-redundant sequences to files
    :type nr_only: bool
    """
    for i, pham in enumerate(phams):
        fasta_path = fasta_dir.joinpath(f"pham_{i+1}.faa")
        if nr_only:
            dump_nr_translations(pham, fasta_path)
        else:
            headers, sequences = list(), list()
            for _, geneid, translation in pham:
                headers.append(geneid), sequences.append(translation)
            write_fasta(headers, sequences, fasta_path)


def map_pangenome(phams, n_genomes):
    """
    Map out which phams belong to the core, soft-core, shell, or cloud
    compartments of the pangenome.

    :param phams: the assembled gene phamilies
    :type phams: list[Pham]
    :param n_genomes: the number of genomes examined
    :type n_genomes: int
    :return: pangenome_map
    """
    pangenome_map = {"core": [], "soft-core": [], "shell": [], "cloud": []}

    for i, pham in enumerate(phams):
        phamid = f"pham_{i+1}"
        fraction = float(len(pham.genomes)) / n_genomes
        if fraction >= 0.99:
            pangenome_map["core"].append(phamid)
        elif fraction >= 0.95:
            pangenome_map["soft-core"].append(phamid)
        elif fraction >= 0.15:
            pangenome_map["shell"].append(phamid)
        else:
            pangenome_map["cloud"].append(phamid)

    return pangenome_map


def summarize(pangenome_map, phams, all_genomes, out_dir):
    """
    Create some of the Roary output files.

    :param pangenome_map: the core/soft/shell/cloud mappings
    :type pangenome_map: dict[str[list[Pham]]]
    :param phams: the assembled phams
    :type phams: list[Pham]
    :param all_genomes:
    :type all_genomes: list[str]
    :param out_dir: the directory where Roary-style files can go
    :type out_dir: pathlib.Path
    """
    summary_statistics = out_dir.joinpath("summary_statistics.txt")
    with open(summary_statistics, "w") as summary_writer:
        summary_writer.write(f"Core genes (99% <= strains <= 100%):\t"
                             f"{len(pangenome_map['core'])}\n")
        summary_writer.write(f"Soft core genes (95% <= strains < 99%):\t"
                             f"{len(pangenome_map['soft-core'])}\n")
        summary_writer.write(f"Shell genes (15% <= strains < 95%):\t"
                             f"{len(pangenome_map['shell'])}\n")
        summary_writer.write(f"Cloud genes (0% <= strains < 15%):\t"
                             f"{len(pangenome_map['cloud'])}\n")
        summary_writer.write(f"Total genes (0% <= strains <= 100%):\t"
                             f"{len(phams)}\n")

    core_genes = out_dir.joinpath("core_genes.txt")
    with open(core_genes, "w") as core_writer:
        for core_phamid in pangenome_map['core']:
            core_writer.write(f"{core_phamid}\n")

    soft_genes = out_dir.joinpath("soft_core_genes.txt")
    with open(soft_genes, "w") as soft_writer:
        for soft_phamid in pangenome_map['soft-core']:
            soft_writer.write(f"{soft_phamid}\n")

    shell_genes = out_dir.joinpath("shell_genes.txt")
    with open(shell_genes, "w") as shell_writer:
        for shell_phamid in pangenome_map['shell']:
            shell_writer.write(f"{shell_phamid}\n")

    cloud_genes = out_dir.joinpath("cloud_genes.txt")
    with open(cloud_genes, "w") as cloud_writer:
        for cloud_phamid in pangenome_map['cloud']:
            cloud_writer.write(f"{cloud_phamid}\n")

    strain_genes = out_dir.joinpath("strain_genes.tsv")
    with open(strain_genes, "w") as fh:
        for i, pham in enumerate(phams):
            phamid = f"pham_{i+1}"
            for genome in pham.genomes:
                for _ in pham.get_genome_geneids(genome):
                    fh.write(f"{phamid}\t{genome}\n")

    gene_presence_absence = out_dir.joinpath("gene_presence_absence.csv")
    with open(gene_presence_absence, "w") as roary_writer:
        header = HEADER
        header.extend(all_genomes)
        roary_writer.write(",".join(header)+"\n")
        for i, pham in enumerate(phams):
            phamid = f"pham_{i+1}"
            seq_per_iso = float(len(pham))/len(pham.genomes)
            output_row = [phamid, '', str(len(pham.genomes)),
                          str(len(pham)), str(seq_per_iso),
                          '', '', '', '', '',
                          str(pham.minimum_length()),
                          str(pham.maximum_length()),
                          str(pham.average_length())]
            for genome in all_genomes:
                if pham.in_genome(genome):
                    geneids = pham.get_genome_geneids(genome)
                    geneids = ";".join([x.split()[0] for x in geneids])
                    output_row.append(geneids)
                else:
                    output_row.append('')
            roary_writer.write(",".join(output_row) + "\n")
